#! /usr/bin/env python
""" Captures managed z/VM storage and virtual server info to an output file.

Input:
The script takes the IP address of an ensemble management HMC, a CPC name, 
and an eHMC userid and password as its only required argumnets. 
It also allows the user to input several options.
The available options (in no particular order) are:

    --version                       Print the version of this script and exit
    --help, -h                      Prints help info and exits
    --verbose, -v                   Show virtual servers and error text in stdout
    --name, -n                      Name of specific z/VM to capture. Default: all z/VMs
    --prefix, -p anytext            Filter on names of virtual servers
    --file, -f filename             Output file. Default: script-name.cpc-name

Operation:
The script has the following primary operational steps:
1) Validates input arguments and options.
2) Signs onto the HMC via the SSL API port of 6794 and establishes a session.
3) GETs the ensemble object uri and uses it to GET the ensemble object.
4) GETs the node uri and uses it to get the CPC uri
5) GETs the list of z/VM hosts on this CPC. If the user asked for a 
   specific z/VM, then limit the list to just that z/VM host
6) Iterates through the z/VM host list. For each z/VM, GETs the list
   of storage resources assigned to it and writes those out. Then
   builds a list of storage groups owned by that z/VM, along with a 
   list of the host storage resources belonging to that storage group,
   and writes all that information out. 

   Then it GETs the list of virtual servers being managed for the 
   z/VM instance. This will be filtered to include only virtual 
   servers with the requested name prefix, if any, and 
   sorted by virtual server name. It writes out the name of each 
   virtual server into the output file, along with its gpmp enablement
   status. If the "automate" suite is installed for this ensemble
   then the processor management enablement status will also be
   written out, along with the workloads each virtual server belongs to.
   Then, networking information for the virtual server is written out.
7) Prints out final results to the user and exits the HMC session.

Output:
For each storage device owned by a z/VM instance it writes out the 
device number, volume serial, size, type and name. For fcp/scsi devices
it also writes out info for each path to the devcie (host port wwpn,
controller port wwpn, and lun). It lists all the storage groups 
for each z/VM, and for those storage groupsthat have actual host 
storage resources associated with them it writes out their device 
number and volume serial.

Then it writes out the names of each virtual server to the output file, along with 
their gpmp enablement, processor management enablement status, and the 
workloads it belongs to.

Developed and tested on Linux on System z using python 2.6.5. Also 
tested on Microsoft Windows XP using python 2.7.3.

Copyright 2012 IBM Corporation
"""
import sys, optparse, time, httplib, urllib, json, socket, os, string

# ##########################################################################
# Logon function
# ##########################################################################
def hmc_logon(hmc_ip, userid, password):
    '''Logon to HMC and return the API session header. 
       
       Input is an HMC IPv4 address that is expected to have already
       been validated as a reasonable IP address 
    '''

    headers = {"Content-type":"application/json","Accept":"*/*"}
    # Login credentials live in body
    body = "{\"userid\":\"" + userid + "\",\"password\":\"" + password + "\"}"
 
    conn = httplib.HTTPSConnection(hmc, 6794)
    try:
        conn.request("POST", "/api/session",body,headers)
        response = conn.getresponse()
    except socket.error as errno:
        print "Error attempting to login to HMC at", hmc
        print errno
        print "Ensure IP address is correct and HMC is enabled for API access"
        sys.exit()

    if response.status == 200:
        print "Successfully logged into HMC at IP address", hmc, "at", time.strftime("%m/%d/%Y %H:%M:%S")
    else:
        print "Failed to login to HMC at IP address", hmc
        print "login status", response.status, "login reason", response.reason
        sys.exit()

    # this session header is important.
    apiSessionHeader = response.getheader( "X-API-Session" )
   
    return apiSessionHeader

# ##########################################################################
# Logoff function
# ##########################################################################
def hmc_logoff(hmc_ip, apiSessionHeader):
    '''Delete the IP session to the HMC.
       
       Input is the IPv4 address of the HMC and API session header
    '''
    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)
    conn.request("DELETE", "/api/session/this-session","",headers)
    response = conn.getresponse()

    if response.status != 204:
        print "Failed attempting to delete the HMC session"
        print "delete status", response.status, "delete reason", response.reason
        print "body:", response.read()

    conn.close()

    return

# ##########################################################################
# get_ensemble_uri function to list ensemble to get the ensemble-uri
# ##########################################################################
def get_ensemble_uri(hmc_ip, apiSessionHeader):
    '''Get the ensemble urii and return it to caller.

       Input is the HMC IP address and the session header for the API call.

       If an error is hit, this function will exit the entire program.
    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)
    conn.request("GET", "/api/ensembles", "", headers)
    response = conn.getresponse()

    if response.status != 200:
        print "Attempt to list ensembles failed."
        print "GET status", response.status, "GET reason", response.reason
        hmc_logoff(hmc_ip, apiSessionHeader)
        sys.exit()
    
    #get the json object and convert to a python object
    objs = json.loads(response.read())
    ensemble_uri = objs['ensembles'][0]['object-uri']

    return ensemble_uri

############################################################################
# get_ensemble_obj function to get an ensemble object
############################################################################
def get_ensemble_obj(hmc_ip, apiSessionHeader, ensemble_uri):
    '''Get an ensemble object.

       Input is the HMC IP address, the session header for the API call,
       and the uri for the ensemble object.

       If an error is hit, this function will exit the entire program
    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)
    conn.request("GET", ensemble_uri, "", headers)
    response = conn.getresponse()

    if response.status != 200:
        print "Attempt to get ensemble properties failed"
        print "GET status", response.status, "GET reason", response.reason
        hmc_logoff(hmc_ip, apiSessionHeader)
        sys.exit()

    ensemble_obj = json.loads(response.read())
    return ensemble_obj


###########################################################################
# get_node_uri function Use the ensemble-uri to list ensemble nodes
# and find the node-uri for the inputted CPC name
###########################################################################
def get_node_uri(hmc_ip, apiSessionHeader, ensemble_uri, cpc_name):
    '''Get an ensemble node uri for a given CPC name.

       Input is the HMC IP address, the session header for the API call,
       the ensemble URI for the ensmeble object, and the 
       text name of the target CPC node.

       If an error is hit, this function will exit the entire program
    '''
    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)

    conn.request("GET", ensemble_uri + "/nodes", "", headers)
    response = conn.getresponse()

    if response.status !=200:
        print "Attempt to get node uri failed"
        print "GET status", response.status, "GET reason", response.reason
        sys.exit()

    objs = json.loads(response.read())
    node_list = objs['nodes']

    nodes_found = 0
    for n in node_list:
        if n['name'] == cpc_name:
            node_uri = n['element-uri']
            nodes_found = nodes_found + 1
    if nodes_found == 0:
        print "Requested CPC name not found in ensemble"
        sys.exit()

    return node_uri

###########################################################################
# get_cpc_uri function. Get node properties to get the cpc-uri for the node
###########################################################################
def get_cpc_uri(hmc_ip, apiSessionHeader, node_uri):
    '''Get a cpc uri for a given ensemble node.

       Input is the HMC IP address, the session header for the API call,
       and the node uri.

       If an error is hit, this function will exit the entire program
    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)

    conn.request("GET", node_uri, "", headers)
    response = conn.getresponse()

    if response.status !=200:
        print "Attempt to get node properties failed"
        print "GET status", response.status, "GET reason", response.reason
        sys.exit()

    obj = json.loads(response.read())
    cpc_uri = obj['member']

    return cpc_uri

###########################################################################
# get_zvm_host_list function. Return a list of VM host uris on a CPC
###########################################################################
def get_zvm_host_list(hmc_ip, apiSessionHeader, cpc_uri, zvm_name):
    '''Get a list of zVM host objects on a given CPC.

       Input is the HMC IP address, the session header for the API call,
       the cpc uri, and a specific zVM host name (which might be None).
       If a particular zVM host name was passed as input, then limit
       the returned list to just that host. Otherwise, return all 
       zVM hosts on this cpc.

       If an error is hit, this function will exit the entire program
    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)

    parms = "?type=zvm"
    conn.request("GET", cpc_uri + "/virtualization-hosts" + parms, "", headers)
    response = conn.getresponse()

    if response.status !=200:
         print "Attempt to get zvm virtualization hosts of a CPC failed"
         print "GET status", response.status, "GET reason", response.reason
         sys.exit()

    objs = json.loads(response.read())
    zvm_host_list = objs['virtualization-hosts']

    total_hosts = 0
    if zvm_name != None:                # if specific zVM host requested, remove others from list
        for zvm in zvm_host_list[:]:    # scan a sliced copy of the host list
            if zvm['name'] != zvm_name: 
                zvm_host_list.remove(zvm) # remove non-matching zVM host from list
            else:                         # this is the requested host
                total_hosts = total_hosts + 1 
    else:                               # otherwise, we'll scan all zVM hosts
        total_hosts = len(zvm_host_list)

    if total_hosts == 0:                # Check results and print out messages
        if zvm_name != None:
            print "Requested z/VM host", "'" + zvm_name + "'", "not found in ensemble"
        else:
            print "No z/VM hosts found in ensemble"
        sys.exit()

    return zvm_host_list

###########################################################################
# get_zvm_obj function. Gets a full zvm object
###########################################################################
def get_zvm_obj(hmc_ip, apiSessionHeader, zvm_uri, zvm_name):
    '''Get a zvm object pointed to by a zvm uri

       Input is the HMC IP address, the session header for the API call,
       the zvm's uri, and the zvm's name.

       If an error is hit, the function returns "None" rather than a
       zvm object
    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)

    conn.request("GET", zvm_uri, "", headers)
    response = conn.getresponse()
    if response.status != 200:
        print "\nAttempt to get properties of zvm ", zvm_name, "failed"
        print "GET status", response.status, "GET reason", response.reason
        if response.status == 503:
            response_body = json.loads(response.read())
            print "Reason sub-code", response_body['reason']
            if verbose:
                print "Error message:", response_body['message']
            print "Date and time: ", time.strftime("%m/%d/%Y %H:%M:%S")
        zvm_obj = None
    else:
        zvm_obj = json.loads(response.read())


    return zvm_obj

############################################################################
# get_vs_list functon. GET the virtual server objects for a given z/VM host
############################################################################
def get_vs_list(hmc_ip, apiSessionHeader, zvm, name_prefix):
    '''Get a list of virtual server uris for a zVM host object.

       Input is the HMC IP address, the session header for the API call,
       a zVM host object, and a virtual server name prefix 
       to limit the results to (or None). If a name prefix was passed 
       as input, then limit the returned list to virtual servers beginning 
       with that name. Otherwise, return all virtual servers managed by this z/VM host.

       If an error is hit, this function will exit the entire program
    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)

# Parms included desired virtual server name prefix, if requested
    if name_prefix != None:
        parms = "?name=" + name_prefix + ".*"
    else:
        parms = ""

    conn.request("GET", zvm['object-uri'] + "/virtual-servers" + parms, "", headers)
    response = conn.getresponse()

    if response.status != 200:
        print "Attempt to list virtual servers for z/VM", zvm['name'], "failed"
        print "GET status", response.status, "GET reason", response.reason
        sys.exit()

    objs = json.loads(response.read())
    vs_list = objs['virtual-servers']

    return vs_list


###########################################################################
# get_vo function. Gets a virtual server object
###########################################################################
def get_vo(hmc_ip, apiSessionHeader, vs_uri, vs_name):
    '''Get a virtual server object pointed to by a virtual server uri 

       Input is the HMC IP address, the session header for the API call,
       the virtual server's uri, and the virtual server name

       If an error is hit, the function returns "None" rather than a 
       virtual server object
    '''
   
    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)
    
    conn.request("GET", vs_uri, "", headers)
    response = conn.getresponse()
    if response.status != 200:
        print "\nAttempt to get properties of virtual server ", vs_name, "failed"
        print "GET status", response.status, "GET reason", response.reason
        if response.status == 503:
            response_body = json.loads(response.read())
            print "Reason sub-code", response_body['reason']
            if verbose:
                print "Error message:", response_body['message']
            print "Date and time: ", time.strftime("%m/%d/%Y %H:%M:%S")
        vo = None
    else:
        vo = json.loads(response.read())

    
    return vo

############################################################################
# get_workload_list function. Gets list of workload names for a virtual server
############################################################################
def get_workload_list(hmc_ip, apiSessionHeader, vo):
    '''Get the list of workload names to which the input virtual server belongs.

       Input is the HMC IP address, the session header for the API call,
       and the virtual server object.

       If an error is hit retrieving a given workload object, then
       for that workload the string "workload retrieval error" is returned 
       in the list for that workload rather than an actual workload name. If
       no workloads are found then an empty list is returned.
       
    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)

    workloads = []
    workload_uri_list = vo['workloads']
    for w_uri in workload_uri_list:     # retrieve each workload object pointed to by the vs object
        conn.request("GET", w_uri, "", headers)
        response = conn.getresponse()
        if response.status != 200:  
            print "\nAttempt to get properties of a workload for virtual server", vo['name'], "failed"
            print "GET status", response.status, "GET reason", response.reason
            if verbose:
               print "Full workload uri list for this virtual server: ", vo['workloads']
               print "Workload uri that cannot be found: ", w_uri
            print "Date and time: ", time.strftime("%m/%d/%Y %H:%M:%S")
            workloads.append("workload retrieval error")
        else:                          # add the workload's name to our list
            w = json.loads(response.read())
            workloads.append(w['name'])

    return workloads

############################################################################
# get_stor_list function. Gets a list of storage resources for a z/VM host
############################################################################
def get_stor_list(hmc_ip, apiSessionHeader, zvm):
    '''Get a list of storage resource uris for a zVM host object.

       Input is the HMC IP address, the session header for the API call,
       and a zVM host object. 

       If an error is hit, this function will exit the entire program
    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)
    parms = "?properties=all"

    conn.request("GET", zvm['object-uri'] + "/virtualization-host-storage-resources" + parms, "", headers)
    response = conn.getresponse()

    if response.status != 200:
        print "Attempt to list storage resources for z/VM", zvm['name'], "failed"
        print "GET status", response.status, "GET reason", response.reason
        if response.status == 503:
            response_body = json.loads(response.read())
            print "Reason sub-code", response_body['reason']
            if verbose:
                print "Error message:", response_body['message']
        sys.exit()

    objs = json.loads(response.read())
    stor_list = objs['virtualization-host-storage-resources']

    return stor_list

############################################################################
# get_stor_groups function. Gets a list of storage groups for a z/VM host
############################################################################
def get_stor_groups(hmc_ip, apiSessionHeader, zvm):
    '''Get a list of storage group uris for a zVM host object.

       Input is the HMC IP address, the session header for the API call,
       and a zVM host object.

       If an error is hit, this function will exit the entire program
    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)
    parms = "?properties=all"

    conn.request("GET", zvm['object-uri'] + "/virtualization-host-storage-groups" + parms, "", headers)
    response = conn.getresponse()

    if response.status != 200:
        print "Attempt to list storage groups for z/VM", zvm['name'], "failed"
        print "GET status", response.status, "GET reason", response.reason
        if response.status == 503:
            response_body = json.loads(response.read())
            print "Reason sub-code", response_body['reason']
            if verbose:
                print "Error message:", response_body['message']
        sys.exit()

    objs = json.loads(response.read())
    stor_groups = objs['virtualization-host-storage-groups']

    return stor_groups


############################################################################
# build_stor_group_dict function. Builds a dictionary that containts an etnry
# for each storage group, which contains the name of the storage group 
# and a list of the host storage resources contained in that storage group
############################################################################
def build_stor_group_list(hmc_ip, apiSessionHeader, zvm, stor_list):
    '''Scans the input host storage object list and from it builds a dictionary
       that contains one entry for each storage group found in the inputted
       list. Each entry in the new dictionary contains the name of the storage 
       group and a list of the host storage resources contained in 
       that storage group.

       Input is the HMC IP address, the session header for the API call,
       a z/VM host object and a list of host storage objects.

       If an error is hit, this function will exit the entire program
    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)

    ##########################################################################
    # check each host storage object in the inputted list to see what storage 
    # group it belongs to, if any
    #########################################################################
    stor_group_list = []
    stor_dev_count = 0
    for s in stor_list:
        stor_dev_count = stor_dev_count + 1   # keep track of where we are in loop for printing status
        if (s['type'] == "eckd" or s['type'] == "zmv-fcp") and s['virtualization-host-storage-group'] != None:
            group_uri = s['virtualization-host-storage-group']
        else:            # this isn't a z/VM storage device or isn't a member of a storage group
            continue         # skip to next host storage object in list

        conn.request("GET", group_uri, "", headers)        
        response = conn.getresponse()

        if response.status != 200:
            print "Attempt to get properties for storage group associated with storage device", s['name'], "failed"
            print "GET status", response.status, "GET reason", response.reason
            if response.status == 503:
                response_body = json.loads(response.read())
                print "Reason sub-code", response_body['reason']
                if verbose:
                    print "Error message:", response_body['message']
            sys.exit()

        obj = json.loads(response.read())
        stor_group = obj

        #######################################################################
        # This storage object is in a storage group, and we have retrieved
        # the corresponding storage group object. See whether that storage 
        # group is already represented in the list we are building. If so,
        # then just add this storage object to the list for this storage group.
        # If not, then begin a new list of storage objects associated with 
        # this storage group.
        #######################################################################
        list_found = False
        for sg in stor_group_list:                          # scan list of storage groups found so far...
            if sg['stor_group_name'] == stor_group['name']: # if have started a list of storage objects for 
                                                            #    this storage group
               sg['stor_list'].append(s)                    # then just add this storage object to the list
               list_found = True
        if not list_found:                                  # don't yet have a list for this storage group 
            stor_dict_entry = {'stor_group_name' : stor_group['name'], 'stor_list' : [s]}

            stor_group_list.append(stor_dict_entry)         # begin a list of storage objects for this storage group
                                                            # and add this storage object to that list
        print_progress(stor_dev_count, len(stor_list)) 
    print "100 percent complete" # do outside fo FOR loop to ensure we always end at 100%
    return stor_group_list

################################################################################################
# get_virtual_network_names function. Gets list of virtual network names from a list of network uris
################################################################################################
def get_virtual_network_names(hmc_ip, apiSessionHeader, net_uri_list):
    '''Get list of virtual network names from a list of network uris.

       Input is the HMC IP address, the session header for the API call,
       and a list of virtual network uris (associated with a network adapter).
       Output is a string with the names of the virtual networks.

       If an error is hit retrieving a given network object, then
       for that virtual network the string "*virtual network object retrieval error*"
       is returned in the list for that network rather than an actual network name. 
       If the input list of virtual network uris is null, then an empty
       string is returned. 

    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)

    net_names = ""
    count = 1
    for n_uri in net_uri_list:         # retrieve each network object
        conn.request("GET", n_uri, "", headers)
        response = conn.getresponse()
        if response.status != 200:
            print "\nAttempt to get a virtual network object failed"
            print "GET status", response.status, "GET reason", response.reason
            print "Date and time: ", time.strftime("%m/%d/%Y %H:%M:%S")
            if count == 1:
               net_names = "*virtual network object retrieval error (" + str(response.reason) + ")*"
            else:
               net_names = net_names + ",*virtual network object retrieval error (" + str(response.reason) + ")*"
        else:                          # add the network's name to our list
            n = json.loads(response.read())
            if count == 1:
               net_names = str(n['name'])
            else:          
               net_names = net_names + "," + str(n['name'])
        count = count + 1

    return net_names

################################################################################################
# get_vswitch_name function. Gets name of a virtual switch from a vswtich uri
################################################################################################
def get_vswitch_name(hmc_ip, apiSessionHeader, vswitch_uri):
    '''Get name of a vswitch its uri.

       Input is the HMC IP address, the session header for the API call,
       and the uri for a virtual switch object.
       Output is a string with the name of the virtual switch (or, if
       the input vswitch uri was null, then the function returns 
       "*None defined*").

       If an error is hit retrieving a given vswitch object, then
       the string "*virtual switch object retrieval error*"
       is returned rather than an actual vswitch name.

    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)

    vswitch_name = ""
    if vswitch_uri == None:
       vswitch_name = "*None defined*"
    else:
       conn.request("GET", vswitch_uri, "", headers)
       response = conn.getresponse()
       if response.status != 200:
           print "\nAttempt to get a virtual switch object failed"
           print "GET status", response.status, "GET reason", response.reason
           print "Date and time: ", time.strftime("%m/%d/%Y %H:%M:%S")
           print "vswitch uri=", vswitch_uri 
       else:                          # save the vswitch name
          vsw = json.loads(response.read())
          vswitch_name = str(vsw['name'])
 
    return vswitch_name

###########################################################################
# get_sw_list function. Gets list of virtual switch uris of the type requested
###########################################################################
def get_sw_list(hmc_ip, apiSessionHeader, zvm, sw_type):
    '''Get a list of virtual switch uris of the type requested.

       Input is the HMC IP address, the session header for the API call,
       the zvm host object, and the type of switch desired.
       Output is a list of virtual switch objects of the desired type.
       If no switches of the desired type are found, then a null list
       is returned.

       If an error is hit, this function will exit the entire program.
    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)

    # get the list of virtual switches
    conn.request("GET", zvm['object-uri'] + "/virtual-switches", "", headers)
    response = conn.getresponse()

    if response.status != 200:
        print "Attempt to list virtual switches for z/VM", zvm['name'], "failed"
        print "GET status", response.status, "GET reason", response.reason
        sys.exit()

    objs = json.loads(response.read())
    sw_list = objs['virtual-switches']

    for sw in sw_list[:]:    # scan a sliced copy of the switch list
       if sw['type'] != sw_type:
          sw_list.remove(sw) # remove non-matching switch from list

    return sw_list

###########################################################################
# get_swo_list function. Gets list of virtual switch objects
###########################################################################
def get_swo(hmc_ip, apiSessionHeader, sw_uri, sw_name):
    '''Get a virtual switch object from its uri.

       Input is the HMC IP address, the session header for the API call,
       the switch uri and switch name.
       Output is a list of virtual switch objects of the desired type.

       If an error is hit, a value of "None" is returned.
    '''

    headers = {"Content-type":"application/json","Accept":"*/*","X-API-Session":apiSessionHeader}
    conn = httplib.HTTPSConnection(hmc_ip, 6794)

    conn.request("GET", sw_uri, "", headers)
    response = conn.getresponse()
    if response.status != 200:
        print "\nAttempt to get properties of virtual switch ", sw_name, "failed"
        print "GET status", response.status, "GET reason", response.reason
        if response.status == 503:
            response_body = json.loads(response.read())
            print "Reason sub-code", response_body['reason']
            if verbose:
                print "Error message:", response_body['message']
            print "Date and time: ", time.strftime("%m/%d/%Y %H:%M:%S")
        swo = None
    else:
        swo = json.loads(response.read())


    return swo

##########################################################################################
# get_real_uplknks function. Gets list of a virtual switche's real uplink device numbers
##########################################################################################
def get_real_uplinks(uplink_list):
    '''Get a list of a virtual switche's real uplink device numbers, if any.

       Input is the HMC IP address, the session header for the API call,
       the list of uplink objects.
       Output is a list of zero to three uplink device numbers.
    '''

    uplinks = []
    for upl in uplink_list:
       uplinks.append(upl['device-number'])

    return uplinks

############################################################################
# function to print progress as a percentage 
############################################################################
def print_progress(count, total):
    '''Print progress as a percentage of total items processed. 
    '''

    percent_complete = float(count) / total * 100
    if percent_complete > 10 and percent_complete < 100:  # a cheat for back spacing
        print int(percent_complete), "percent complete",
        print "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b",       # back space
        sys.stdout.flush()
    return


##########################################################################
##########################################################################
# Start of main program. 
##########################################################################
##########################################################################

##########################################################################
# Process and validate input arguments and options
##########################################################################
if __name__ == "__main__":
  try:   # catch IO errors and keyboard interrupt
    usage = "usage: %prog [options] HMCIP CPCNAME USERID PASSWORD"
    version = "%prog 1.0"
    p = optparse.OptionParser(usage=usage, version=version)
    p.add_option('--verbose', '-v',
                 action='store_true',
                 help='include details and error text in stdout')
    p.add_option('--name', '-n',
                 action='store',
                 help='Name of z/VM to capture. Default: all z/VMs')
    p.add_option('--prefix', '-p',
                 action='store',
                 help='Virtual server name prefix to capture')
    p.add_option('--file', '-f',
                 action='store',
                 help='Output file. Default: script-name.cpc-name')
    options, args = p.parse_args()

    verbose         = options.verbose
    zvm_name        = options.name
    name_prefix     = options.prefix

    if options.file == None:
        f_list = string.split(__file__, "/")  # split file path into dirs and filename
        file_name = f_list[len(f_list)-1]     # file name without the path
        s_list = string.split(file_name, ".") # separate main script name fron ".py"
        script_name = s_list[0]
        output_filename = script_name + "." + args[1]  # default
    else:
        output_filename = options.file

    if len(args) == 0:
        p.error("An HMC IP address, CPC name, HMC userid and password are required")
    elif len(args) > 4:
        p.error("Too many arguments")
    else:
    # test to see if the inputted HMC address is a valid IP address
        if len(args[0].split(".")) != 4:         # first test of IP address
            p.error("Invalid IP address")
        try:
            socket.inet_aton(args[0])            # second test of IP address
            hmc = args[0]                        # save HMC IP address
        except:
            p.error("Invalid IP address")

        cpc_name = args[1]                       # save CPC name
        hmc_userid = args[2]                     # save HMC userid
        hmc_password = args[3]                   # save HMC password

##############################################################################
# Begin main processing
#############################################################################

    # logon to HMC
    apiSessionHeader = hmc_logon(hmc, hmc_userid, hmc_password)

    # get the ensemble ufi
    ensemble_uri = get_ensemble_uri(hmc, apiSessionHeader)

    ###########################################################################
    # use ensemble-uri to get ensemble properties and find management 
    # enablement level
    ###########################################################################

    # get the ensemble object
    ensemble_obj = get_ensemble_obj(hmc, apiSessionHeader, ensemble_uri)

    print "\nEnsemble name:", ensemble_obj['name']
    print "Management enablement level:", ensemble_obj['management-enablement-level']
    print "CPU performance management for z/VM enabled at ensemble level:", ensemble_obj['cpu-perf-mgmt-enabled-zvm']

    if ensemble_obj['management-enablement-level'] != 'automate':
        print "\nManagement enablement level is not 'automate'"
        print "Enabling processor management and defining custom workloads are not supported."


    ############################################################################
    # Use the ensmble uri and the input node name to get the node uri, 
    # abd then use the node_uri to get the cpc_uri
    ############################################################################
    node_uri = get_node_uri(hmc, apiSessionHeader, ensemble_uri, cpc_name)
    cpc_uri  = get_cpc_uri(hmc, apiSessionHeader, node_uri)

    ############################################################################
    # Get ths list of target z/VM host uris
    ############################################################################
    zvm_host_list = get_zvm_host_list(hmc, apiSessionHeader, cpc_uri, zvm_name)

    ############################################################################
    # Print out some informational messages to the user
    ############################################################################
    print "Storage, network, and managed virtual servers for these z/VM hosts on CPC", "'" + cpc_name + "'","will be captured:"
    for zvm in zvm_host_list:
        print "     " + zvm['name']

    if name_prefix != None:
        print "Only virtual servers whose names start with", "'" + name_prefix + "'", "will be captured"

    ###########################################################################
    # Iterate through z/VM hosts
    ###########################################################################
    vs_grand_total = 0
    total_error_count = 0

    with open(output_filename, 'w') as f: # open the output file. Close when do
        # Write a global header to the output file 
        f.write("Output of script " + sys.argv[0] + "\n")  # file name of this script
        f.write("Capture of z/VM resources defined to IBM zEnterprise Unified Resource Manager (zManager)\n")
        f.write("Ensemble managment enablement level: " + ensemble_obj['management-enablement-level'] + "\n")
        if  ensemble_obj['management-enablement-level'] == "automate":
            f.write("CPU performance management for z/VM enabled at ensemble level: " + str(ensemble_obj['cpu-perf-mgmt-enabled-zvm']) + "\n\n")
        else:
            f.write("Workload and CPU performance management information is not relevant so will not be captured.\n\n")


        for zvm in zvm_host_list:
            ###########################################################
            # get a list of all storage resources for a particular z/VM
            # and write key info on each to file
            ##########################################################
            print "\nCapturing storage devices for z/VM", "'" + zvm['name'] + "'", "..."
            stor_list = get_stor_list(hmc, apiSessionHeader, zvm)
            stor_list.sort(key=lambda k: k['device-number'])

            f.write("Storage devices defined to zManager for z/VM host " + str(zvm['name']) + " on host " + cpc_name + " dumped on " + time.strftime("%m/%d/%Y %H:%M:%S") + "\n")

            stor_count = 0
            no_sg_group = 0
            for s in stor_list: 
                if s['virtualization-host-storage-group'] == None:
                    f.write("*")                # flag devices that are not part of any storage group
                    no_sg_group = no_sg_group + 1

                f.write("devno=" + str(s['device-number']).ljust(10))
                f.write("volser=" + str(s['volume-serial-number']).ljust(10))

                f.write(str(s['allocation-units']) + "=" + str(s['size']).ljust(10))
                f.write("type=" + str(s['type']).ljust(10) + "name=" + str(s['name']) + "\n")
                if s['type'] == "zvm-fcp":  # extra processing for fcp/scsi devices
                    for p in s['paths']:    # write out info on wwpns and lun for each path to the device
                        f.write("                " + "hp-wwpn= " + str(p['host-port-wwpn']).ljust(10))
                        f.write("cp-wwpn= " + str(p['controller-port-wwpn']).ljust(10))
                        f.write("lun= " + str(p['lun']) + "\n")
                stor_count = stor_count + 1
            if no_sg_group > 0:
                f.write("\n(Devices above flagged with a '*' do not belong to any storage group)\n")
            f.write("\n")
            f.write("Total of " + str(stor_count) + " Storage resources defined to zManager found for z/VM " + str(zvm['name']) + "\n")
            f.write("\n")

           ###########################################################
            # get a list of all storage groups for a particular z/VM
            # and write key info to file
            ###########################################################
            print "Capturing storage groups for z/VM", "'" + zvm['name'] + "'", "..."
            stor_groups = get_stor_groups(hmc, apiSessionHeader, zvm)
            stor_groups.sort(key=lambda k: k['name'])

            f.write("Storage groups defined to zManager for z/VM host " + str(zvm['name']) + " on host " + cpc_name + " dumped on " + time.strftime("%m/%d/%Y %H:%M:%S") + "\n")

            stor_group_count = 0
            for g in stor_groups:
                f.write("group name=" + str(g['name']).ljust(10) + "\n")
                stor_group_count = stor_group_count + 1
            f.write("\n")
            f.write("Total of " + str(stor_group_count) + " Storage groups defined to zManager found for z/VM " + str(zvm['name']) + "\n")

            #############################################################
            # Build a list of storage object associated with each storage
            # group, then write out the results to file
            #############################################################
            stor_group_list = build_stor_group_list(hmc, apiSessionHeader, zvm, stor_list) 
            f.write(str(len(stor_group_list)))  # write out number of storage groups that have actual devices
            if len(stor_group_list) == 1:
                f.write(" storage group has storage devices defined to it via zManager\n")
            else:
                f.write(" storage groups have storage devices defined to them via zManager\n")

            for g in stor_group_list:
                f.write("\nStorage devices defined to zManager belonging to storage group " + str(g['stor_group_name']) + "\n")
                for s in g['stor_list']:    # write out the list of storage devices in this storage group
                    f.write("devno=" + str(s['device-number']).ljust(10))
                    f.write("volser=" + str(s['volume-serial-number']).ljust(10) + "\n") 

            f.write ("\n")

            ###########################################################
            # Now get a list of all virtual switches for this z/VM host
            # and write out info to the output file.
            ###########################################################

            # get the full zzm obejct, since one of its fields will be needed later
            zvm_obj = get_zvm_obj(hmc, apiSessionHeader, zvm['object-uri'], zvm['name'])
            
            ########################################
            # iterate through the iedn vswitch list
            ########################################

            # Print some info out to user
            print "Capturing IEDN virtual switches for z/VM", "'" + zvm['name'] + "'", "..."

            # initialize variables
            iedn_sw_count = 0
            error_count = 0
            error_sw_list = []

            # get a list of iedn vswitches
            iedn_sw_list = get_sw_list(hmc, apiSessionHeader, zvm, "iedn")

            # write out a header for iedn vswitches
            f.write("IEDN virtual switches for z/VM host " + str(zvm['name']) + " on host " + cpc_name + " dumped on " + time.strftime("%m/%d/%Y %H:%M:%S"))

            for sw in iedn_sw_list:
               iedn_sw_count = iedn_sw_count + 1   # count of all iedn vswitches processed for this z/VM

               # Get the actual virtual switch object
               swo = get_swo(hmc, apiSessionHeader, sw['element-uri'], sw['name'])
               if swo == None:                     # error in retrieving switch object
                  error_count = error_count + 1
                  error_sw_list.append(sw['name']) # add to list of virtual servers we could not retrieve
                  continue

               # get list of real uplink device numbers, if any
               real_uplinks = get_real_uplinks(swo['real-uplinks'])
               

               # write out basic iedn vswitch info
               f.write("\nname=" + str(swo['name']).ljust(10) + "layer-mode=" + str(swo['layer-mode']).ljust(5) + \
                       "queue-size=" + str(swo['queue-size']).ljust(3) + "ip-timeout=" + str(swo['ip-timeout']).ljust(4) + \
                       "connect-uplinks=" + str(swo['is-connect-uplinks']).ljust(7))

               if real_uplinks == []:
                  f.write("real uplinks=None\n")
               else:
                  f.write("real uplinks(dev.port)=[")
                  for u in real_uplinks:
                     f.write(" " + str(u)+ " ")        
                  f.write("]\n")

               # write out attributes that may or may not exist for this vswitch
               attr_written = False
               if zvm_obj['is-bridge-capable']:                # this z/VM is bridge capable?
                  f.write("mtu-size-enforcement=".rjust(36) + str(swo['mtu-size-enforcement']).ljust(14))
                  attr_written = True
                  if swo['mtu-size-enforcement'] == "user-defined":
                     f.write("mtu-size=" + str(swo['mtu-size']).ljust(7))
                  if swo['layer-mode'] == "eth":               # vswitch is layer2?
                     if swo['bridge-device-number'] != None:   # vswitch has a valid bridge device?
                        f.write("bridge-device-number=" + str(swo['bridge-device-number']).ljust(6))
                        f.write("bridge-port-type=" + str(swo['bridge-value-type']).ljust(11))

               if swo['layer-mode'] == "ip":
                  if attr_written:          # have already indented on a new line
                     f.write("router=" + str(swo['router']))
                  else:                     # on a new line, so must indent
                     f.write("router=".rjust(22) + str(swo['router']))
                     attr_written = True

               if attr_written:
                  f.write("\n")  

            f.write("\nTotal IEDN vswitches found = " + str(iedn_sw_count) + "\n")
            if error_count == 0:
               f.write("No found IEDN vswitches had errors preventing their retrieval\n\n")
            else:
                f.write("\nThe following IEDN vswitches had errors and could not be retrieved:\n")
                for bad_sw in error_sw_list:
                    f.write(str(bad_sw) + "\n")
                f.write("\n")


            ########################################
            # iterate through the qdio vswitch list
            ########################################

            # Print some info out to user
            print "Capturing QDIO virtual switches for z/VM", "'" + zvm['name'] + "'", "..."


            # initialize variables
            qdio_sw_count = 0
            error_count = 0
            error_sw_list = []

            # get a list of qdio vswitches
            qdio_sw_list = get_sw_list(hmc, apiSessionHeader, zvm, "qdio")

            # write out a header for qdio vswitches
            f.write("QDIO virtual switches for z/VM host " + str(zvm['name']) + " on host " + cpc_name + " dumped on " + time.strftime("%m/%d/%Y %H:%M:%S"))

            for sw in qdio_sw_list:
               qdio_sw_count = qdio_sw_count + 1   # count of all iedn vswitches processed for this z/VM

               # Get the actual virtual switch object
               swo = get_swo(hmc, apiSessionHeader, sw['element-uri'], sw['name'])
               if swo == None:                     # error in retrieving switch object
                  error_count = error_count + 1
                  error_sw_list.append(sw['name']) # add to list of virtual servers we could not retrieve
                  continue

               # get list of real uplink device numbers, if any
               real_uplinks = get_real_uplinks(swo['real-uplinks'])


               # write out basic iedn vswitch info
               f.write("\nname=" + str(swo['name']).ljust(10) + "layer-mode=" + str(swo['layer-mode']).ljust(5) + \
                       "queue-size=" + str(swo['queue-size']).ljust(3) + "ip-timeout=" + str(swo['ip-timeout']).ljust(4) + \
                       "connect-uplinks=" + str(swo['is-connect-uplinks']).ljust(7))

               if real_uplinks == []:
                  f.write("real uplinks=None\n")
               else:
                  f.write("real uplinks(dev.port)=[")
                  for u in real_uplinks:
                     f.write(" " + str(u)+ " ")
                  f.write("]\n")

               # write out vlan-related info (mostly only if switch is vlan aware)
               f.write("vlan-aware=".rjust(26) + str(swo['is-vlan-aware']).ljust(7))
               if swo['is-vlan-aware']:
                  f.write("vlan-id=" + str(swo['vlan-id']).ljust(6) + "vlan-port-type=" + str(swo['vlan-port-type']).ljust(8) + \
                          "vlan-native-id=" + str(swo['vlan-native-id']).ljust(6) + "gvrp-enabled=" + str(swo['is-gvrp-enabled']).ljust(7))
               f.write("\n")

               # write out attributes that may or may not exist for this vswitch
               attr_written = False
               if zvm_obj['is-bridge-capable']:                # this z/VM is bridge capable?
                  f.write("mtu-size-enforcement=".rjust(36) + str(swo['mtu-size-enforcement']).ljust(14))
                  attr_written = True
                  if swo['mtu-size-enforcement'] == "user-defined":
                     f.write("mtu-size=" + str(swo['mtu-size']).ljust(7))
                  if swo['layer-mode'] == "eth":               # vswitch is layer2?
                     if swo['bridge-device-number'] != None:   # vswitch has a valid bridge device?
                        f.write("bridge-device-number=" + str(swo['bridge-device-number']).ljust(6))
                        f.write("bridge-port-type=" + str(swo['bridge-value-type']).ljust(11))

               if swo['layer-mode'] == "ip":
                  if attr_written:          # have already indented on a new line
                     f.write("router=" + str(swo['router']))
                  else:                     # on a new line, so must indent
                     f.write("router=".rjust(22) + str(swo['router']))
                     attr_written = True

               if attr_written:
                  f.write("\n")

            f.write("\nTotal QDIO vswitches found = " + str(qdio_sw_count) + "\n")
            if error_count == 0:
               f.write("No found QDIO vswitches had errors preventing their retrieval\n\n")
            else:
                f.write("\nThe following QDIO vswitches had errors and could not be retrieved:\n")
                for bad_sw in error_sw_list:
                    f.write(str(bad_sw) + "\n")
                f.write("\n")


            ###########################################################
            # now move on to process virtual servers
            ###########################################################
            error_vs_list = []  # keep track of names of virtual servers we could not retrieve
            # get a list of virtual servers for this z/VM host abd sort it
            vs_list = get_vs_list(hmc, apiSessionHeader, zvm, name_prefix)
            vs_list.sort(key=lambda k: k['name'])

            # Initialize some variables
            vs_count = 0
            error_count = 0
            total_vs = len(vs_list)

            # Print some info out to user
            print "Capturing virtual servers for z/VM", "'" + zvm['name'] + "'", "..."
            print "   ",   # indent progress ticker

            ##############################################################
            # Write header info out to file for each z/VM processed
            ##############################################################
            f.write("Managed virtual servers for z/VM host " + str(zvm['name']) + " on host " + cpc_name + " dumped on " + time.strftime("%m/%d/%Y %H:%M:%S") + "\n")
            if name_prefix != None:
                f.write("Only virtual servers whose names start with '" + name_prefix + "' will be dumped.\n")

            ##############################################################
            # Iterate through the list of virtual server uris
            ##############################################################
            for v in vs_list:
                vs_count = vs_count + 1         # count of all virtual servers processed for this z/VM

                #################################################################
                # Get the actual virtual server object
                #################################################################
                vo = get_vo(hmc, apiSessionHeader, v['object-uri'], v['name'])
                if vo == None:                      # Error in retrieiving vs object
                    error_count = error_count + 1
                    error_vs_list.append(v['name']) # add to list of virtual servers we could not retrieve
                    continue

                # write out the virtual server info, one vs per line
                f.write("\nname=" + str(vo['name']).ljust(20) + "gpmp=" + str(vo['gpmp-support-enabled']).ljust(7))
                if  ensemble_obj['management-enablement-level'] == "automate": # only relevant for automate suite
                    f.write("cpu_mgmt=" + str(vo['cpu-perf-mgmt-enabled']) + "\n")
                
                    # get list of workload names to which this virtual server belongs
                    workloads = get_workload_list(hmc, apiSessionHeader, vo)
                    # write out the workload names, indented on a new line
                    f.write("workloads= ".rjust(48))
                    if workloads == []:         # should never happen
                        f.write("This virtual server does not belong to any workloads\n")
                    else:
                        w_count = 1
                        for w in workloads:
                            f.write(w)
                            if w_count < len(workloads):
                                f.write(", ")
                            else:
                                f.write("\n")
                            w_count = w_count + 1
                else:
                    f.write("\n")

                na_list = vo['network-adapters']   # get list of network adapters for this virtual server
                if len(na_list) == 0:
                   f.write("No network adapters defined\n".rjust(53))
                else:                              # write out networking info for this virtual server
                   for na in na_list:
                      f.write("vNIC=".rjust(30))
                      f.write(str(na['virtual-device-address']).ljust(6))
                      f.write(" type=" + str(na['type']).ljust(5))
                      if na['type'] != "rmc":      #rmc devices do not have an "interface-type' field
                         f.write(" interface type=" + str(na['interface-type']).ljust(13) + "\n")
                         f.write("".rjust(36))  # indent on a new line
                         if na['interface-type'] == "virtual-iedn" or na['interface-type'] == "virtual-qdio":
                            if na['interface-type'] == None:
                               f.write(" port mode N/A")
                            else:
                               f.write(" port mode=" + str(na['port-mode']))
                               f.write(" vswitch=" + get_vswitch_name(hmc, apiSessionHeader, na['switch-uri']))
                         if na['interface-type'] == "physical-iedn" or \
                            na['interface-type'] == "physical-qdio" or \
                            na['interface-type'] == "physical-iqdn" :
                            f.write(" rNIC=" + str(na['real-device-address']).ljust(5))
                         if na['interface-type'] == "virtual-qdio":
                            f.write("  vlan ids=" + str(na['vlan-ids'])) 
                         if na['interface-type'] == "virtual-iedn" or  \
                            na['interface-type'] == "physical-iedn" :
                            f.write("  virtual networks=" + get_virtual_network_names(hmc, apiSessionHeader, na['network-uris']))
                      f.write("\n")
               
                # print progress as a percentage
                print_progress(vs_count, total_vs)

            # processing for a zVM host is done, write out some final markers
            print "100 percent complete" # do outside fo FOR loop to ensure we always end at 100%
        
            if vs_count == 0:
                f.write("\nNo virtual servers satisfying the request for this z/VM were found.\n")
            if error_count == 0:
                f.write("\nNo virtual servers had errors preventing their retrieval.\n")
            else:
                f.write("\nThe following virtual servers had errors and could not be retrieved:\n")
                for bad_vs in error_vs_list:
                    f.write(str(bad_vs) + "\n")
            f.write("*********************************************************\n\n") # end of file output fot this z/VM

            ############################################################################
            # Print the results to screen if requested by user
            ############################################################################
            if verbose:
                print "\nThe following operating virtual servers were found for z/VM", "'" + zvm['name'] + "'", ":"
                if len(vs_list) == 0:
                   print "<None>"
                else:
                    for v in vs_list:
                        print v['name']
                print "\n", error_count, "found virtual servers could not be captured for z/VM", "'" + zvm['name'] + "'"

            ############################################################################
            # Keep track of grand totals for final messates to user
            ###########################################################################
            vs_grand_total = vs_grand_total + vs_count
            total_error_count = total_error_count + error_count

    ###########################################################################
    # Done with main processing. Report grand totals to user
    ###########################################################################
    print "\n\nTotal of", vs_grand_total - total_error_count, "virtual servers written to file", output_filename
    print "Total of", total_error_count, "virtual servers had errors and could not be captured\n"

    # ##########################################################################
    # Now we'll logoff and a print a completion notice
    # ##########################################################################

    hmc_logoff(hmc, apiSessionHeader)

    print "Script completed at", time.strftime("%H:%M:%S")

  ##############################################################################
  # Catch errors from "try" statement at beginning of main program
  ##############################################################################
  except KeyboardInterrupt:
    print "\nScript execution cancelled by user"
    sys.exit()
  except IOError:
    print "\nError opening or writing to output file", output_filename
    print "Execution aborted"
    sys.exit()

